import io
import sys
import os

import generalized_multi
from generalized_multi.storage import Storage, \
     update_itemset, ItemsetInfoForExhaustiveSearch
from generalized_multi.combinations import *

def delete_keys(dct, to_delete):
    n_of_deletion = 0
    for k in to_delete:
        n_of_deletion += 1
        del dct[k]
    return n_of_deletion

def generalize_unfrequent_itemsets(storage_unit, current_iteration, generalizer, minsup, apply_generalization = True):
    """
    If an itemset has support  < minsup, it has to be removed, and its generalization
    marked to be kept.

    An itemset frequent which is not generalized has to be kept.

    If apply_generalization is true, then frequent generalized itemsets which contain 
    at least 1 unfrequent itemset are included in the output
    """
    m = storage_unit.get_map_for_length(current_iteration)
    to_delete = []

    for itemset in m.iterkeys():
        # Test if itemset is rare
        itemset_info = m[itemset]
        s, generalized, keep = itemset_info

        include_generalized = False

        if s >= minsup and not generalized:
            m[itemset].set(keep = True)

        if s < minsup:
            # if itemset is rare, it must be deleted from the set of itemset
            to_delete.append(itemset)
            if s > 0:
                # if itemset is rare, its generalization has to be marked if apply_generalization is True
                include_generalized = apply_generalization

        if include_generalized:
            for new_itemset in generalizer.generalize_itemset(itemset):
                info = m.get(new_itemset, None)
                if info is None:
                    continue
                m[new_itemset].set(keep = True)

    delete_keys(m, to_delete)


def ensure_subsets_inclusion(storage_unit, length_of_itemsets, i_to_check_for_subs_inclusion, output_writer):
    next_checks = []
    length = length_of_itemsets  - 1

    if length <= 0:
        return

    l_map = storage_unit.get_map_for_length(length)
    for itemset in i_to_check_for_subs_inclusion:
        for subitemset in k_subsets(itemset, length):
            info =  l_map.get(subitemset, None)
            # If info is None, then the itemset and its subsets 
            # have already been written
            if info is not None:
                # No support checking is required...
                output_writer.write_itemset_info(subitemset, info)
                next_checks.append(subitemset)
                del l_map[subitemset]

    return ensure_subsets_inclusion(storage_unit, length, next_checks, output_writer)

# def is_a_subset(itemset, transaction, length):
#     """
#     """
#     tags = [x[0] for x in itemset]
#     if length != len(set(tags)):
#         return False
#     return itemset.issubset(transaction)

def main(dataset_reader, generalizer, minsup, info_logger, output_writer,
         candidate_generator, no_gc = False, ensure_subsets = False, basic_pruning = True, *args, **kwargs):
    """
    If basic_pruning is True, then generalized itemsets which contain only frequent itemsets
    are discarded from the output

    """

    ensure_subsets = ensure_subsets and basic_pruning
    if no_gc:
        import gc
        gc.disable()

    storage_unit = Storage(itemset_info_class = ItemsetInfoForExhaustiveSearch)

    iteration = 0
    completed = False

    while not completed:
        iteration += 1
        info_logger.info("Iterazione %s" % iteration, )
        default_spec = ItemsetInfoForExhaustiveSearch(support=1, generalized=False, keep=False)
        default_gen = ItemsetInfoForExhaustiveSearch(support=1, generalized=True, keep = not basic_pruning)
        if iteration == 1:
            for item in dataset_reader.get_items_iterator():
                update_itemset(storage_unit, frozenset( (item,) ), default_spec, create_if_not_found = True)
                for i in generalizer.get_all_generalizations(item):
                    i = frozenset((i,))
                    if len(i) != 0:
                        update_itemset(storage_unit, i, default_gen, create_if_not_found = True)

        else:
            it = dataset_reader.get_extended_transactions_iterator()
            m = storage_unit.get_map_for_length(iteration)

            # update support of itemsets in storage
            for itemset in it:
                for k in m.iterkeys():
                    #if is_a_subset(k, itemset, iteration):
                    if k.issubset(itemset):
                        m[k].support += 1

        # process unfrequent itemsets: discard ALL unfrequent itemsets
        # mark generalizations of unfrequent if basic_pruning is True
        generalize_unfrequent_itemsets(storage_unit, iteration, generalizer, minsup, 
                                       apply_generalization = basic_pruning)
        
        # candidate generation
        info_logger.info("Candidati")
        has_new_itemsets = candidate_generator(storage_unit,iteration+1)
        completed = not has_new_itemsets
        
        info_logger.info("Scrittura degli %s-itemset frequenti" % (iteration,))
        # frequent itemsets not generalized
        # and generalized itemsets marked as keep are now useless
        m = storage_unit.get_map_for_length(iteration)
        to_delete = []
        i_to_check_for_subs_inclusion = []
        for itemset in m.iterkeys():
            info = m[itemset]
            support, generalized, keep = info
            if (not generalized) or (generalized and keep):
                output_writer.write_itemset_info(itemset, info)
                to_delete.append(itemset)
                if generalized and ensure_subsets:
                    i_to_check_for_subs_inclusion.append(itemset)
                    
        ensure_subsets_inclusion(storage_unit, iteration, i_to_check_for_subs_inclusion, output_writer)

        for k in to_delete:
            del m[k]

    if no_gc:
        gc.collect()
    
    output_writer.append(dataset_reader.number_of_transactions)

