import io
import sys
import os

import generalized_multi
from generalized_multi.storage import Storage, \
     ItemsetInfoWithExclusionSet, update_itemset, ItemsetStorage
from generalized_multi.combinations import *

def delete_keys(dct, to_delete):
    for k in to_delete:
        del dct[k]

def generalize_unfrequent_itemsets(source, destination, current_iteration, generalizer, minsup):
    """
    If k = current_iteration
    Unfrequent k-itemsets in source are generalized and stored in destination
    """
    source_map = source.get_map_for_length(current_iteration)
    destination_map = destination.get_map_for_length(current_iteration)
    default_spec = ItemsetInfoWithExclusionSet(support=0,exclusion_set = True)

    unfrequent_accumulator = []
    _append = unfrequent_accumulator.append
    # generating generalized
    for itemset, itemset_info in source_map.iteritems():
        s = itemset_info.support
        if s >= minsup:
            continue
        _append(itemset)
        if s == 0:
            continue
        for new_itemset_l1 in generalizer.generalize_itemset(itemset):
            # if level-1 is missing, then all its generalizations have to be generated
            if update_itemset(destination, new_itemset_l1, default_spec, create_if_not_found = True):
                continue
            for new_itemset_upper in generalizer.generalize_itemset(new_itemset_l1, all_levels=True):
                update_itemset(destination, new_itemset_upper, default_spec, create_if_not_found = True)

    # Delete keys
    delete_keys(source_map, unfrequent_accumulator)

def yield_generalization_info(spec_itemsets_map, gen_itemsets_map, generalizer):
    """
    For each itemset in itemsets_map, yield its info and the info of its generalizations
    """
    for itemset, itemset_info in spec_itemsets_map.iteritems():
        for new_itemset in generalizer.generalize_itemset(itemset, all_levels = True):
            info = gen_itemsets_map.get(new_itemset, None)
            if info is not None:
                yield itemset, itemset_info, new_itemset, info

def initialize_s_set(generalized_itemsets, specific_itemsets, length, generalizer):
    """
    Optimization. To reduce the cost of support counting while scanning the dataset.
    Exclusion set is assigned to each generalization
    """    
    specific_itemsets_map = specific_itemsets.get_map_for_length(length)
    generalized_itemsets_map = generalized_itemsets.get_map_for_length(length)
    # Update support of generalized itemsets
    support_correction = 0
    for i, i_info, j, j_info in yield_generalization_info(specific_itemsets_map, generalized_itemsets_map, generalizer):
        j_info.exclusion_set.add(i)
        support_correction = i_info.support
        j_info.s_support += support_correction
        j_info.support -= support_correction

def compute_opportunistic(itemset, itemset_map, pruning_condition, minsup, itemsets_to_delete):
    """
    Compute opportunistic support of itemset, traversing itemsets in its
    exclusion list. If an itemset in its exclusion list has support < minsup, it
    must be excluded from exclusion list.
    """
    itemset_info = itemset_map.get(itemset, None)

    if itemset_info is None:
        return 0

    s, exclusion_set, opportunistic_computed, s_support = itemset_info

    if pruning_condition(itemset_info, minsup):
        itemsets_to_delete.add(itemset)
        return 0

    if opportunistic_computed:
        return s

    s_support = sum(compute_opportunistic(x, itemset_map, pruning_condition, \
                                          minsup, itemsets_to_delete) for x in exclusion_set)
    itemset_info.support -= s_support
    itemset_info.s_support += s_support
    exclusion_set.difference_update(itemsets_to_delete)
    itemset_info.opportunistic_computed = True
    
    if pruning_condition(itemset_info, minsup):
        itemsets_to_delete.add(itemset)
        return 0
    
    return itemset_info.support

def opportunistic_pruning_condition(itemset_info, minsup):
    return itemset_info.support < minsup

def basic_pruning_condition(itemset_info, minsup):
    return (itemset_info.support == 0) or (itemset_info.support + itemset_info.s_support) < minsup
    
def prune_generalized_itemsets(storage_unit, length, generalizer, minsup, basic_pruning = False):
    """
    Compute opportunistic support considering levels of generalization!
    If support is already lower than minsup, no further computation is needed, 
    since it can only decrease...
    """
    
    unfrequent_accumulator = []    
    destination_map = storage_unit.get_map_for_length(length)

    for i, i_info, j, j_info in \
        yield_generalization_info(destination_map, destination_map, generalizer):
        if j_info.support >= minsup and i_info.support >= minsup and i != j:
            j_info.exclusion_set.add(i)
            
    to_remove_from_exclusion = set([])
    add_unfrequent = unfrequent_accumulator.append

    if basic_pruning:
        pruning_condition = basic_pruning_condition
    else:
        pruning_condition = opportunistic_pruning_condition
        
    for itemset in destination_map.iterkeys():
        to_remove_from_exclusion.clear()
        new_support = compute_opportunistic(itemset, destination_map, pruning_condition, minsup, to_remove_from_exclusion)
        info = destination_map[itemset]
        if (info.support + info.s_support) < minsup:
            add_unfrequent(itemset)
            
    # Delete keys
    delete_keys(destination_map, unfrequent_accumulator)

def prune_unfrequent(storage, length, minsup):
    unfrequent_accumulator = []
    destination_map = storage.get_map_for_length(length)

    for itemset, itemset_info in destination_map.iteritems():
        if itemset_info.support < minsup:
            unfrequent_accumulator.append(itemset)

    for itemset in unfrequent_accumulator:
        del destination_map[itemset]
            
def dump_itemsets(itemsets, writer, minsup, delete = True):
    """
    Write itemsets using writer.
    if delete is True, then they are deleted
    """
    for itemset,info in itemsets.iteritems():
        if info.support >= minsup:
            writer(itemset, info, itemsets)
    if delete:
        itemsets.clear()

def scan_dataset_and_update_support(dataset_reader, *args):
    """
    args contains maps of itemsets

    Scan the dataset updating the support of itemsets contained in maps 
    passed as *args
    """
    it = dataset_reader.get_extended_transactions_iterator()

    for transaction in it:
        for it_map in args:
            for k in it_map.iterkeys():
                if k.issubset(transaction):
                    it_map[k].support += 1
    

def force_subsets_inclusion(itemsets_to_search, forced_subsets_su, length, *args):
    """
    itemsets_to_search is a list of itemsets whose subsets must be included
    forced_subsets_su is the StorageUnit which will host subitemset to include
    length is the length of itemsets in itemsets_to_search

    optional positional arguments *args contain StorageUnit which are scanned to discover if
    a particular subitemset is already known
    """

    if length <= 1:
        return

    target = length - 1
    maps = [su.get_map_for_length(target) for su in args]
    default_spec = ItemsetInfoWithExclusionSet(support=0,exclusion_set = False)

    for itemset in itemsets_to_search:
        for subitemset in k_subsets(itemset, target):
            if any( (subitemset in m) for m in maps ):
                continue
            update_itemset(forced_subsets_su, subitemset, default_spec, create_if_not_found = True)

    map_new_forced = forced_subsets_su.get_map_for_length(target)
    force_subsets_inclusion(map_new_forced.iterkeys(), forced_subsets_su, target, *args)


def sac_version(dataset_reader, generalizer, minsup, info_logger, output_writer,
                    candidate_generator, no_gc = False, **kwargs):

    if no_gc:
        import gc
        gc.disable()

    specific_itemsets = Storage(itemset_info_class = ItemsetInfoWithExclusionSet)
    generalized_itemsets = Storage(itemset_info_class = ItemsetInfoWithExclusionSet, generalized = True)
    
    completed = False
    iteration = 0

    while not completed:
        iteration += 1
        info_logger.info("Iteration %s" % iteration, )

        if iteration == 1:
            default_spec = ItemsetInfoWithExclusionSet(support=1,exclusion_set = False)
            for item in dataset_reader.get_items_iterator():
                update_itemset(specific_itemsets, frozenset( (item,) ), 
                               default_spec, create_if_not_found = True)
        else:
            # update support of itemsets in storage
            scan_dataset_and_update_support(dataset_reader, specific_itemsets.get_map_for_length(iteration))

        generalize_unfrequent_itemsets(specific_itemsets, generalized_itemsets,
                                       iteration, generalizer, minsup)

        lst = []
        s, m = specific_itemsets.get_map_for_length(iteration), generalized_itemsets.get_map_for_length(iteration)
        for k in m.iterkeys():
            if k in s:
                lst.append(k)
        for k in lst:
            del m[k]
        
        info_logger.info("Supporto generalizzati" )

        scan_dataset_and_update_support(dataset_reader, generalized_itemsets.get_map_for_length(iteration))
        prune_unfrequent(generalized_itemsets, iteration, minsup)
            
        info_logger.info("Generazione candidati" )
        candidate_generator(specific_itemsets, specific_itemsets, specific_itemsets, iteration+1)
        candidate_generator(specific_itemsets, generalized_itemsets, specific_itemsets, iteration+1)
        candidate_generator(generalized_itemsets, generalized_itemsets, specific_itemsets, iteration+1)

        dump_itemsets(specific_itemsets.get_map_for_length(iteration),
                      output_writer.write_itemset_info, minsup)
        dump_itemsets(specific_itemsets.get_map_for_length(iteration),
                      output_writer.write_itemset_info, minsup)
        completed = not specific_itemsets.has_itemsets_of_length(iteration+1)
    
    if no_gc:
        gc.collect()

    output_writer.append(dataset_reader.number_of_transactions)


def genio_post_pruning(dataset_reader, generalizer, minsup, info_logger, output_writer,
                       candidate_generator, no_gc = False, ensure_subsets = False, basic_pruning = False, \
                       no_exclusion_set = False, **kwargs):

    def fill_s_set(iteration, storage_units):
        # Fill the exclusion set
        for level in sorted(storage_units.keys()):
            to_delete = []
            m = storage_units[level].get_map_for_length(iteration)
            
            for itemset, itemset_info in m.iteritems():

                if (itemset_info.support + itemset_info.s_support) < minsup:
                    to_delete.append(itemset)
                    continue

                if itemset_info.support < minsup:
                    continue
                
                for higher_itemset, h_level in generalizer.generalize_itemset(itemset, \
                                                                              all_levels = True, yield_level = True):
                    h_info = storage_units[h_level].get_map_for_length(iteration)[higher_itemset]
                                        
                    h_info.exclusion_set.add(itemset)
                    h_info.support -= itemset_info.support
                    h_info.s_support += itemset_info.support

            for i in to_delete:
                del m[i]

    if no_exclusion_set:
        raise ValueError, "I need the S set..."

    if no_gc:
        import gc
        gc.disable()

    storage_units = {} #Storage(itemset_info_class = ItemsetInfoWithExclusionSet, generalized = True)
    default_info = ItemsetInfoWithExclusionSet(support=1,exclusion_set = True)

    # First Iteration
    iteration = 1
    completed = False

    while not completed:
    
        info_logger.info("Iteration %s" % iteration, )

        if iteration == 1:
            # Support count
            for transaction in dataset_reader.get_extended_transactions_iterator():
                for item in transaction:
                    level = generalizer.get_item_level(item)
                    if level not in storage_units:
                        storage_units[level] = Storage(itemset_info_class = ItemsetInfoWithExclusionSet, \
                                                       generalized = (level != 0))
                    update_itemset(storage_units[level], frozenset( (item,) ), default_info, create_if_not_found = True)
        else:
            for transaction in dataset_reader.get_extended_transactions_iterator():
                for level, storage_unit in storage_units.iteritems():
                    m = storage_unit.get_map_for_length(iteration)
                    for itemset in m.iterkeys():
                        if itemset.issubset(transaction):
                            m[itemset].support += 1

    
        fill_s_set(iteration, storage_units)

        # candidate generation
        has_new_itemsets = False
        levels = sorted(storage_units.keys())
        for n, level_a in enumerate(levels):
            for level_b in levels[n:]:
                new_itemsets = candidate_generator(storage_units[level_a], storage_units[level_b], \
                                                   storage_units[level_b], iteration + 1)
                has_new_itemsets = has_new_itemsets or new_itemsets

        # Print
        for k, storage_unit in storage_units.iteritems():
            dump_itemsets(storage_unit.get_map_for_length(iteration),
                            output_writer.write_itemset_info, minsup)

        
        iteration += 1
        completed = not has_new_itemsets
            
    output_writer.append(dataset_reader.number_of_transactions)
    
    if no_gc:
        gc.collect()

        
def genio(dataset_reader, generalizer, minsup, info_logger, output_writer,
          candidate_generator, no_gc = False, ensure_subsets = False, basic_pruning = False, \
          no_exclusion_set = False, **kwargs):

    """
    If basic_pruning is True, then a generalized itemset is kept if its opportunistic support is > 0.

    if no_exclusion_set is True, then an exclusion set is not created. Note that if the hierarchy
    has two levels, this is the similar to set basic_pruning to True.
    """
    
    if no_gc:
        import gc
        gc.disable()

    specific_itemsets = Storage(itemset_info_class = ItemsetInfoWithExclusionSet)
    generalized_itemsets = Storage(itemset_info_class = ItemsetInfoWithExclusionSet, generalized = True)
    forced_subsets = Storage(itemset_info_class = ItemsetInfoWithExclusionSet)

    completed = False
    iteration = 0

    while not completed:

        iteration += 1
        info_logger.info("Iteration %s" % iteration, )

        # On the first iteration, count support of specific itemsets
        if iteration == 1:
            default_spec = ItemsetInfoWithExclusionSet(support=1,exclusion_set = False)
            for item in dataset_reader.get_items_iterator():
                update_itemset(specific_itemsets, frozenset( (item,) ), 
                               default_spec, create_if_not_found = True)

        else:
            # On iteration k > 1, count support of specific k-itemsets and (k-1)-itemsets
            maps = [specific_itemsets.get_map_for_length(iteration), 
                    generalized_itemsets.get_map_for_length(iteration - 1)]

            # update support of itemsets in storage
            scan_dataset_and_update_support(dataset_reader, *maps)


        # once the support of generalized itemsets has been computed, the *opportunistic* support
        # must be computed and unfrequent itemsets dropped
        if no_exclusion_set:
            prune_unfrequent(generalized_itemsets, iteration-1, minsup)
        else:
            prune_generalized_itemsets(generalized_itemsets, iteration-1, generalizer, minsup, \
                                       basic_pruning = basic_pruning)

        # generalize and delete level-0 unfrequent itemsets.
        generalize_unfrequent_itemsets(specific_itemsets, generalized_itemsets,
                                       iteration, generalizer, minsup)
        
        info_logger.info("Generation of candidates")
        # specific candidates of length iteration+1 are generated using 
        # specific itemsets of length iteration.
        candidate_generator(specific_itemsets, specific_itemsets, specific_itemsets, iteration+1)
        # generalized candidates of length iteration are generated using 
        # specific and generalized itemsets of length iteration-1
        candidate_generator(specific_itemsets, generalized_itemsets, generalized_itemsets, iteration)
        candidate_generator(generalized_itemsets, generalized_itemsets, generalized_itemsets, iteration)
        
        # Once generalized itemset of length iteration are generated
        # the support of frequent specific itemsets must be subtracted from every generalization
        # and the exclusion list must be filled
        if not no_exclusion_set:
            initialize_s_set(generalized_itemsets, specific_itemsets, iteration, generalizer)

        # If iteration is k
        # specific (k-1)-itemsets and generalized (k-1)-itemsets can be written on disk
        if iteration > 1:
            target = iteration - 1
            info_logger.info("Writing frequent %s-itemset" % (target,))
            dump_itemsets(specific_itemsets.get_map_for_length(target),
                                   output_writer.write_itemset_info, minsup, delete = not ensure_subsets)
            # If all the subsets of frequent itemsets must be kept, then frequent itemsets
            # of level > 1 are not discarded
            dump_itemsets(generalized_itemsets.get_map_for_length(target),
                                   output_writer.write_itemset_info, minsup, delete = not ensure_subsets)

            if ensure_subsets and target > 1:
                # if k-itemsets with level >= 1 and k > 1 have been generated
                # then subsets inclusion must be ensured
                m = generalized_itemsets.get_map_for_length(target)
                force_subsets_inclusion(m.iterkeys(), forced_subsets, target, generalized_itemsets, specific_itemsets)

        completed = not (specific_itemsets.has_itemsets_of_length(iteration+1) or 
                         generalized_itemsets.has_itemsets_of_length(iteration))

    if no_gc:
        gc.collect()

    info_logger.info("Writing frequent %s-itemset" % (iteration,))
    for storage_u in (specific_itemsets, generalized_itemsets):
        itemsets = storage_u.get_map_for_length(iteration)
        dump_itemsets(itemsets, output_writer.write_itemset_info, minsup)

    if ensure_subsets:
        # Count support for itemsets whose inclusion has been forced
        scan_dataset_and_update_support(dataset_reader, *forced_subsets.get_maps())
        for itemsets in forced_subsets.get_maps():
            dump_itemsets(itemsets, output_writer.write_itemset_info, minsup)

    output_writer.append(dataset_reader.number_of_transactions)
