"""
Association rules extractor
"""


import io
from combinations import k_subsets
import re

def get_repr(left, right, tail, *meas):
    """
    return a string with the form

    left => right tail (meas)
    """
    def _to_str(lst, sep = ' '):
        return sep.join(str(x) for x in lst)
    def _write_itemset(it):
        return ('%s:%s' % (y[0], y[1]) for y in it)
    left = _to_str(_write_itemset(left))
    right = _to_str(_write_itemset(right))
    meas = _to_str(meas, sep = ',')
    return '%s => %s %s (%s)' % (left, right, tail, meas)

def get_key(sort_by):
    keys = {'support': 3, 'confidence':4, 'lift':5}
    def _get_k(rule):
        return rule[keys[sort_by]]
    return _get_k

class ItemsetInfo(object):
    __slots__ = ["support", "s_text", "s_support"]
    def __init__(self, support = 0, s_text="", s_support=0):
        self.support = support
        self.s_text = s_text
        self.s_support = s_support
        
    def __repr__(self):
        toks = [str(self.__class__)] + [ "%s:%s" % (x,str(getattr(self, x)))\
                                             for x in ItemsetInfo.__slots__]
        return ' '.join(toks)

def decode_itemset_trad(itemset, rest):
    support = int(rest.strip('()'))
    itemset = frozenset(itemset)
    return itemset, ItemsetInfo(support = support)

match_num = re.compile('\d+')
def decode_itemset_opp(itemset, rest):
    def extract_num(st):
        m = match_num.search(st)
        return int(m.group())
    s_text, measures = tuple(rest.strip().split('(', 1))

    measures_toks = [x.strip() for x in measures.split(',')]
    support = extract_num(measures_toks[0])
    s_support = 0
    if len(measures_toks) > 1:
        s_support = extract_num(measures_toks[1])

    itemset = frozenset(itemset)
    info = ItemsetInfo(support=support, s_text=s_text, s_support = s_support)
    return itemset, info

def get_itemset_decoder(opportunistic = False):
    if opportunistic:
        return '\\(', decode_itemset_opp
    else:
        return '(', decode_itemset_trad

def all_subitemsets(itemset):
    for size in xrange(1, len(itemset)):
        for subitemset in k_subsets(itemset, size):
            yield subitemset

def iterate_logging_perc(iterator, info_logger):
    for n, data in enumerate(iterator()):
        if n % 100 == 0:
            info_logger.info('%s itemsets have been processed' % (n, ))
        yield data

def get_subitem_support(info, superitem_info, n_tr = 1):
    """
    Return the correct support of subitem with respect to a superset of it,
    by considering both the s_supports

    n_tr is the number of transactions (needed to compute the relative support)
    """
    fixed = info.support + info.s_support - superitem_info.s_support
    return 1.0 * fixed / n_tr

def main(dataset_encoded, minconf, info_logger, output_file_obj, sort_by = "lift", 
         ascending=False, opportunistic = False):
    """
    sort_by can be "support", "confidence", "lift"
    """
    dataset_reader = io.DatasetReader(dataset_encoded, None, info_logger.info)
    n_transactions = 0
    itemsets = {}
    ch_to_ignore, decode_itemset = get_itemset_decoder(opportunistic = opportunistic)

    for itemset, rest in dataset_reader.transactions_iterator(ignore_after = ch_to_ignore):
        itemset, info = decode_itemset(itemset, rest)
        if len(itemset) == 0:
            n_transactions = info.support
        else:
            itemsets[itemset] = info

    out_list = []
    n_of_itemsets = dataset_reader.number_of_transactions
    
    for itemset,info in iterate_logging_perc(itemsets.iteritems, info_logger):
        support = 1.0*info.support / n_transactions
        s_support = info.s_support
        s_text = info.s_text


        for candidate_body in all_subitemsets(itemset):
            try:
                info_body = itemsets[candidate_body]
                support_body = get_subitem_support(info_body, info, n_tr = n_transactions)

                confidence = 1.0 * support / support_body
                if confidence < minconf:
                    continue
                head_itemset = itemset.difference(candidate_body)
                info_head = itemsets[head_itemset]
                support_head = get_subitem_support(info_head, info, n_tr = n_transactions)
                lift = confidence / support_head
                out_list.append( (candidate_body, head_itemset, s_text, support, confidence,lift) )
            except KeyError:
                info_logger.warning('Some itemsets are missing... try re-running the fim with -m option')

    del itemsets

    info_logger.info('Sorting by %s' % (sort_by,))
    out_list.sort(key=get_key(sort_by), reverse = not ascending)

    for rule in out_list:
        output_file_obj.write('%s\n' % (get_repr(*rule)))
