import sys
import os
import os.path
import shelve
from itertools import islice
import logging

class Logger(object):
    __console_set = False
    def __init__(self, name, filename):
        logging.basicConfig(level=logging.DEBUG,
                            format='[%(asctime)s] %(levelname)-8s %(message)s',
                            datefmt='%m-%d %H:%M',
                            filename=filename,
                            filemode='a')
        if not Logger.__console_set:
            console = logging.StreamHandler()
            console.setLevel(logging.WARNING)
            formatter = logging.Formatter('[%(asctime)s] %(levelname)-8s %(message)s')
            console.setFormatter(formatter)
            # add the handler to the root logger
            logging.getLogger('').addHandler(console)
            Logger.__console_set = True

        self.logger = logging
    
    def __getattr__(self, name):
        return getattr(self.logger, name)  

def load_item(item):
    t = tuple([int(x) for x in item.split(':')])
    return t

class DatasetReader(object):
    def __init__(self, filename, generalizer, log_func, log_every_n_lines=1000):
        self.filename = filename
        self.generalizer = generalizer
        self.number_of_transactions = 0
        self.log_func = log_func
        self.log_every_n_lines = log_every_n_lines

    def _ping(self):
        if self.number_of_transactions % self.log_every_n_lines == 0:
            self.log_func("Reading line %s" % (self.number_of_transactions,))
            
    def _string_lines_iterator(self):
        """
        Iterate on file yielding text lines
        """
        self.number_of_transactions = 0
        for line in file(self.filename, 'r', 256*1024):
            self._ping()
            l = line.strip()
            if l:
                self.number_of_transactions += 1
                yield l.split()

    def transactions_iterator(self, ignore_after = None):
        """
        Iterates over lines yielding list of items ( pairs of ints )
        """
        def _iterates_no_skip():
            for line in self._string_lines_iterator():
                yield [ load_item(x) for x in line]

        def _iterates_skipping():
            for tokens in self._string_lines_iterator():
                l = len(tokens)
                left_part = []
                while len(tokens) > 0:
                    if any( x in tokens[0] for x in ignore_after):
                        break
                    left_part.append(load_item(tokens.pop(0)))
                yield left_part, ' '.join(tokens)

        if ignore_after is None:
            return _iterates_no_skip()
        else:
            return _iterates_skipping()
        
    def get_items_iterator(self):
        """
        tuples of ints!
        """
        for itemset in self.get_extended_transactions_iterator(extend = False):
            for item in itemset:
                yield item

    
    def get_extended_transactions_iterator(self, extend = True):
        """
        yield frozensets!
        """
        for it in self.transactions_iterator():
            if extend:
                extended = it[:]
                for i in it:
                    extended.extend(self.generalizer.get_all_generalizations(i))
                yield frozenset(extended)
            else:
                yield frozenset(it)

class OutputWriter(object):
    def __init__(self, fileobj):
        self.obj = fileobj

    def build_output_st(self, itemset, *args):
        return ' '.join("%s:%s" % item for item in itemset )

    def write_itemset_info(self, itemset, itemset_info, *args):
        """
        Return the itemset
        """
        s = self.build_output_st(itemset, *args)
        s =  "%s (%s)\n" % (s, itemset_info.support + getattr(itemset_info, "s_support", 0))
        self.obj.write(s)
        
    def append(self, n_of_transactions):
        s = "(%s)\n" % (n_of_transactions,)
        self.obj.write(s)
   

class OutputWriterWithExclusionSet(OutputWriter):
    def build_output_st(self, itemset, itemsets_collection):
        itemset_info = itemsets_collection.get(itemset, None)
        s_itemset = super(OutputWriterWithExclusionSet, self).build_output_st(itemset)

        if itemset_info is not None and itemset_info.exclusion_set is not None:
            excl = "\ { %s } " % \
                (' | '.join(self.build_output_st(it, itemsets_collection) for it in itemset_info.exclusion_set),)
        else:
            excl = ""
        return "%s %s" % (s_itemset, excl)

    def write_itemset_info(self, itemset, itemset_info, *args):
        """
        Return the itemset
        """
        s = self.build_output_st(itemset, *args)
        s =  "%s (%s, s(S)=%s)\n" % (s, itemset_info.support, itemset_info.s_support)
        self.obj.write(s)

class OutputWriterWithExclusionSetCompact(OutputWriterWithExclusionSet):
    def get_global_exclusion_set(self, itemset, itemsets_collection, father_exclusion_set,
                                 include_itemsets_excluded_by_itemset = False):
        """
        Fill father_exclusion_set with all the itemsets in the exclusion sets of
        itemsets excluded by itemset
        """
        itemset_info = itemsets_collection.get(itemset, None)
        if itemset_info is None:
            return 
        exclusion_set = itemset_info.exclusion_set
        if exclusion_set is None:
            return
        if include_itemsets_excluded_by_itemset:
            father_exclusion_set.update(exclusion_set)
        for x in itemset_info.exclusion_set:
            self.get_global_exclusion_set(x,itemsets_collection, father_exclusion_set,
                                          include_itemsets_excluded_by_itemset = True)
        
    def build_output_st(self, itemset, itemsets_collection):
        def _to_st(itemset):
            return super(OutputWriterWithExclusionSet, self).build_output_st(itemset)

        itemset_info = itemsets_collection.get(itemset, None)
        s_itemset = _to_st(itemset)

        if itemset_info is not None and itemset_info.exclusion_set is not None:
            tot_exclusion_set = set([])
            self.get_global_exclusion_set(itemset, itemsets_collection, tot_exclusion_set)
            excl = "\ { %s } " % \
                (' | '.join(_to_st(i) for i in itemset_info.exclusion_set if i not in tot_exclusion_set),)
        else:
            excl = ""
        return "%s %s" % (s_itemset, excl)

    def write_itemset_info(self, itemset, itemset_info, *args):
        """
        Return the itemset
        """
        s = self.build_output_st(itemset, *args)
        s =  "%s (%s, s(S)=%s)\n" % (s, itemset_info.support, itemset_info.s_support)
        self.obj.write(s)

def get_output_writer(fobj, print_excluded_strategy = "ignore"):
    if print_excluded_strategy == "ignore":
        return OutputWriter(fobj)
    if print_excluded_strategy == "extended":
        return OutputWriterWithExclusionSet(fobj)
    if print_excluded_strategy == "compact":
        return OutputWriterWithExclusionSetCompact(fobj)
    raise ValueError, "Print excluded strategy can be 'ignore', 'extended' or 'compact' "


import tempfile
import shutil

def wrap_passing_info_logger_and_tempdir(log_file, func, *args, **kwargs):
    # Start logging
    info_logger = Logger(str(os.getpid()), log_file)
    info_logger.info("Start: pid=%s" % (os.getpid() ))

    # Temporary directory
    temporary_dir = tempfile.mkdtemp('', "%s-" % (os.getpid(), ))
    info_logger.info("Temporary directory %s created" % (temporary_dir, ))

    try:
        return func(info_logger, temporary_dir, *args, **kwargs)

    except Exception, e:
        info_logger.exception("an exception occurred")
        
    finally:
        info_logger.info('Removing temporary directory %s' % (temporary_dir,) )
        shutil.rmtree(temporary_dir)

    return -1
