import anydbm
import shelve
import UserDict
import copy

"""

Format: dictionary

keys -> serialized itemsets
values -> (support, generalized, keep)
"""

class ItemsetInfo(object):
    """
    Class which inherits from ItemsetInfo must have a support attribute
    """
    __slots__ = []

    """
    Info and basic operations for each itemset
    """
    def __init__(self, **kwargs):
        raise NotImplementedError

    def __iter__(self):
        for slot in self.__slots__:
            yield getattr(self, slot)

    def __repr__(self):
        l = []
        for k in  self.__slots__:
            l.append("%s:%s" % (k, getattr(self, k)))
        return ', '.join(l)
 
    def set(self, **kwargs):
        """
        Update info using keyword arguments
        """
        for k,v in kwargs.iteritems():
            setattr(self, k, v)

    def update_in_place(self, other_info):
        """
        Update in place this info object
        """
        raise NotImplementedError

    def itemset_info_combinator(self,other):
        """
        Create a new info object referring to the self join of this object whith other.

        """
        raise NotImplementedError


class ItemsetInfoForExhaustiveSearch(ItemsetInfo):
    __slots__ = ["support", "generalized", "keep"]

    def __init__(self, support=1, generalized=False, keep=False):
        self.support = support 
        self.generalized = generalized
        self.keep = keep

    def update_in_place(self, other_info):
        """
        Update in place.... sum supports and do logical or
        on other attributes
        """
        self.support += other_info.support
        self.generalized = self.generalized or other_info.generalized
        self.keep = self.keep or other_info.keep

    def itemset_info_combinator(self,other):
        generalized = self.generalized or other.generalized
        #keep = a.keep and b.keep   -- FIX!
        keep = (self.keep and self.generalized) or (other.keep and other.generalized)
        return ItemsetInfoForExhaustiveSearch(support = 0, generalized = generalized, keep = keep)

class ItemsetInfoWithExclusionSet(ItemsetInfo):
    __slots__ = ["support", "exclusion_set", "opportunistic_computed", "s_support"]

    def __init__(self, support=0, exclusion_set=False, s_support = 0):
        """
        exclusion_set = False force opportunistic_computed to be True
        during initialization
        """
        self.support = support
        self.s_support = s_support
        if exclusion_set:
            self.opportunistic_computed = False
            self.exclusion_set = set([])
        else:
            self.exclusion_set = None
            self.opportunistic_computed = True
             
    def update_in_place(self, other_info):
        """
        Sum supports and do logical and
        add other_info.exclusion_set
        """
        self.support += other_info.support
        if self.exclusion_set is not None and other_info.exclusion_set is not None:
            self.exclusion_set.update(other_info.exclusion_set)

    def itemset_info_combinator(self,other):
        if (self.exclusion_set is not None) or (other.exclusion_set is not None):
            add_exc_set = True
        else:
            add_exc_set = False
        return ItemsetInfoWithExclusionSet(support = 0, exclusion_set = add_exc_set)

class ItemsetStorage(dict):
    def __init__(self, id=None, **kwargs):
        super(ItemsetStorage, self).__init__(**kwargs)

class Storage(object):
    """
    Organizer of itemsets... 1 map for each length
    """
    def __init__(self, map_class = ItemsetStorage, itemset_info_class = ItemsetInfo, generalized = False):
        self.stores = {}
        self.map_class = map_class
        self.itemset_info_class = itemset_info_class
        self.generalized = generalized

    def get_map_for_length(self, l):
        return self.stores.get(l,  self.map_class(id=l) )

    def get_maps(self):
        return self.stores.values()

    def __iter__(self):
        for k,v in self.stores.iteritems():
            yield k,v

    def has_itemsets_of_length(self, l):
        return len(self.stores.get(l, ())) > 0

def update_itemset(storage_unit, itemset, info, create_if_not_found = True):
    """
    Insert itemset into storage_unit
    Return True if itemset was already present
    """
    l = len(itemset)
    if l == 0:
        return
    
    _map = storage_unit.stores.setdefault( l, storage_unit.map_class(id=l) )
    
    old_info = _map.get(itemset, None)
    if old_info is not None:
        old_info.update_in_place(info)

    elif create_if_not_found:
        new_info = copy.deepcopy(info)
        _map[itemset] = new_info

    return (old_info is not None)
