from storage import update_itemset, ItemsetInfo

def self_join_traditional(p, k, candidate, tags, length):
    """
    A standard self join between p,k
    candidate will host the new candidate
    tags is a list whose length is length which is passed as parameter to avoid multiple allocations
    """
    current_item = 0
    matches = 0
    matches_required = length-2
    found_a_match = False

    for item in p:
        candidate[current_item] = item
        current_item += 1
    candidate[current_item] = None

    new_elem = False
    for item in k:
        # Ho un match
        if item in candidate:
            matches += 1
        else:
            # Item diverso e nuovo tag
            if new_elem is True:
                return False
            candidate[current_item] = item
            new_elem = True
    if matches > matches_required:
        return False
    return True

def self_join(p, k, candidate, tags, length):
    """
    A standard self join between p,k
    candidate will host the new candidate
    tags is a list whose length is length which is passed as parameter to avoid multiple allocations
    """
    current_item = 0
    matches = 0
    matches_required = length-2
    found_a_match = False

    for item in p:
        candidate[current_item] = item
        tags[current_item] = item[0]
        current_item += 1
    candidate[current_item] = None
    tags[current_item] = None

    new_elem = False
    for item in k:
        # Ho un match
        if item in candidate:
            matches += 1
        else:
            t,v = item
            # Item diverso ma tag gia' inserito
            if t in tags: 
                return False
            else: 
                # Item diverso e nuovo tag
                if new_elem is True:
                    return False
                candidate[current_item] = item
                tags[current_item] = t
                new_elem = True
    if matches > matches_required:
        return False
    return True

def candidate_generator(store_a, store_b, length, join = self_join_traditional):
    """
    Scan store_a, joining (length-1)-itemsets with itemsets in store_b
    yielding candidates

    Reference implementation. (C implementation is obsolete)
    """
    
    left_map_to_use = store_a.get_map_for_length(length-1)
    right_map_to_use = store_b.get_map_for_length(length-1)
    
    has_new_itemsets = False
    required_matches = length - 2
    new_candidate = [None for x in range(length)]
    new_tags = [None for x in range(length)]

    for p, p_details in left_map_to_use.iteritems():
        s1 = p_details.support

        for k, k_details in right_map_to_use.iteritems():
            s2 = k_details.support

            if not join(p, k, new_candidate, new_tags, length):
                continue

            new_candidate_set = frozenset(new_candidate)
            info = p_details.itemset_info_combinator(k_details)

            yield new_candidate_set, info

def traditional_candidate_generation(destination_store, generator, *args):
    """
    Store the new candidate
    """        
    has_new_candidates = False
    for new_candidate, info in generator(*args):
        has_new_candidates = True
        update_itemset(destination_store, new_candidate, info, create_if_not_found=True)
    return has_new_candidates

def produce_generalized_version_also_candidate_generation(generalizer, destination_store, generator, *args):
    """
    For each new candidate, produce also all its generalizations
    """
    store_a_g, store_b_g = args[0].generalized, args[1].generalized

    if (store_a_g and store_b_g) or not (store_a_g and store_b_g):
        return traditional_candidate_generation(destination_store, generator, *args)
    
    has_new_itemsets = False
    for new_candidate, info in generator(*args):
        has_new_itemsets = True
        update_itemset(destination_store, new_candidate, info, create_if_not_found=True)
        
        for new_upper_candidate in generalizer.generalize_itemset(new_candidate):
            has_new_itemsets = True
            update_itemset(destination_store, new_upper_candidate, info, create_if_not_found=True)
    return has_new_itemsets
    
def well_formed_candidate_generation(generalizer, destination_store, generator, *args):
    """
    Makes a new candidate well-formed, using generalizer, and store it
    """
    has_new_itemsets = False
    for new_candidate, info in generator(*args):
        for new_wf_candidate in generalizer.get_well_formed_versions(new_candidate):
            has_new_itemsets = True
            update_itemset(destination_store, new_wf_candidate, info, create_if_not_found=True)
    return has_new_itemsets

def all_levels_wf_candidate_generation(generalizer, destination_store, generator, *args):
    """
    Return candidates and all upper level well-formed generalizations of each new candidate
    """
    has_new_itemsets = 0
    for new_candidate, info in generator(*args):
        
        level, max_level_can_rise, wf = generalizer.is_well_formed(new_candidate)
        
        if wf:
            has_new_itemsets += 1
            update_itemset(destination_store, new_candidate, info, create_if_not_found=True)

        if level == 0:
            continue
        
        for upper_level in generalizer.generalize_itemset(new_candidate, all_levels = True):
            has_new_itemsets += 1
            update_itemset(destination_store, upper_level, info, create_if_not_found=True)
            
    return has_new_itemsets > 0

def get_candidate_generation_procedure(strategy_name, generalizer, extended=False, exploit_structured = True):
    """
    strategy can be 'traditional' or 'well-formed'

    if extended is False, build a function which receives three parameters
      - store, an instance of class Storage which contains the itemset of level length - 1
      - length, i.e. the length of the candidate to generate
      - minsup, i.e. the minimum support

    if extended is True, build a function which receives five parameters
      - store_a, an instance of class Storage which contains the itemset of level length - 1
      - store_b, an instance of class Storage which contains the itemset of level length - 1 to join
                 with store_a
      - store_c, an instance of class Storage to host the itemset of level length
      - length, i.e. the length of the candidate to generate
    """
    
    if exploit_structured:
        def _generator(*args, **kwargs):
            return candidate_generator(join = self_join, *args, **kwargs)
    else:
        def _generator(*args, **kwargs):
            return candidate_generator(join = self_join_traditional, *args, **kwargs)
    
    if strategy_name == "traditional":
        _candidate_generation = traditional_candidate_generation
    elif strategy_name == 'traditional_all_generalizations':
        def _candidate_generation(*args):
            return produce_generalized_version_also_candidate_generation(generalizer, *args)
    elif strategy_name == "well-formed":
        def _candidate_generation(*args):
            return well_formed_candidate_generation(generalizer, *args)
    elif strategy_name == "genio":
        def _candidate_generation(*args):
            return all_levels_wf_candidate_generation(generalizer, *args)
    else:
        return None

    if extended:
        def wrap_candidate_generation(store_a, store_b, store_c, length):
            return _candidate_generation(store_c, _generator, store_a, store_b, length)
    else:
        def wrap_candidate_generation(store, length):
            return _candidate_generation(store, _generator, store, store, length)
 
    return wrap_candidate_generation
