#!/usr/bin/python

"""
Utilities for encoding/decoding dataset, itemsets, association rules
"""

import sys
import shelve
from optparse import OptionParser

class Encoder(object):
    def __init__(self, filename, sep=':'):
        self.counter = 0
        self.filename_syms = "%s_syms" % ( filename, )
        self.filename_codes = "%s_codes" % ( filename, )
        self.dbm_syms = None
        self.dbm_codes = None
        self.separator = sep
        
    def _encode_symbol(self, sym):
        if sym not in self.dbm_syms:
            code = self.counter
            self.dbm_syms[sym] = code
            self.dbm_codes[str(code)] = sym
            self.counter += 1
        return self.dbm_syms[sym]

    def encode(self, item):
        attribute, value = tuple(item.split(self.separator))
        return ( self._encode_symbol(attribute), self._encode_symbol(value))

    def open(self, reload_old_data=False):
        if reload_old_data:
            mode = 'w'
        else:
            mode = 'n'
        self.dbm_syms = shelve.open(self.filename_syms, mode, protocol=2)
        self.dbm_codes = shelve.open(self.filename_codes, mode, protocol=2)
        if 'code' in self.dbm_codes:
            self.counter = self.dbm_codes['code']

    def close(self):
        self.dbm_syms.close()
        self.dbm_syms = None
        self.dbm_codes['code'] = self.counter
        self.dbm_codes.close()
        self.dbm_codes = None

    def _decode_symbol(self, code):
        return self.dbm_codes[str(code)]

    def decode(self, item):
        attribute, item = item.split(self.separator)
        value_sym = self._decode_symbol(item)
        attr_sym = self._decode_symbol(attribute)
        return (attr_sym, value_sym)


def serialize_in_ascii(encoded_item, sep=':'):
    return sep.join( str(x) for x in encoded_item )

def encode_file(encoder, filename, newfilename, item_separator, skip_last_element=False):
    out = file(newfilename, 'w' )

    last = None
    if skip_last_element:
        last = -1

    for line in file(filename):
        line = line.strip()
        if not line:
            continue
        items = [x.strip() for x in line.split(item_separator)]

        tail = ''
        if last is not None:
            tail = items[last]

        encoded = []
        for x in items[:last]:
            x = x.strip()
            try:
                encoded.append(serialize_in_ascii(encoder(x)))
            except (KeyError, ValueError):
                encoded.append(x)

        newline = item_separator.join(encoded)
        newline = '%s %s' % (newline,tail,)
        out.write('%s\n' % (newline.strip(),))
    out.close()


file_types = {'transactions': 0, 'rules': 1, 'itemsets': 2, 'ar': 3}

def main(filename, dbm_prefix,output_file, decode=False, file_type=0, load_old_data=False):
    seps = [' ', '=>', ' ', ' ']
    sep = seps[file_type]

    if decode:
        encoder = Encoder(dbm_prefix)
        encoder.open(reload_old_data = True)
        encode_file(encoder.decode, filename, output_file, sep, skip_last_element=(file_type>=2))
    else:
        encoder = Encoder(dbm_prefix)
        encoder.open(reload_old_data=load_old_data)
        encode_file(encoder.encode, filename, output_file, sep, skip_last_element=(file_type>=2))
        encoder.close()

