from io import load_item
from combinations import *

class Generalizer(object):

    def _level_update(self, target_item, candidate_new_level):
        level, generalization = self.map.setdefault(target_item, (candidate_new_level, None))
        if candidate_new_level > level:
            # aggiorno il livello per questo item
            self.map[target_item] = candidate_new_level, generalization
            # lancio l'aggiornamento sui padri
            self._level_update(generalization, candidate_new_level + 1)

    def get_item_level(self, item):
        return self.map.get(item, (0,None))[0]

    def __init__(self, filename=None):
        """
        self.map is a dict. Key: item, Value: (level, upper level transformations)
        """
        self.map = {}
        if filename is not None:
            for line in file(filename):
                line = line.strip()
                if line:
                    tokens = [ x.strip() for x in line.split('=>')]
                    left, right = load_item(tokens[0]), load_item(tokens[1])
                    # Handle left part
                    child_level, to_ignore = self.map.get(left, (0, None ) )
                    self.map[left] = child_level, right
                    # handle right part ....  increase the level
                    self._level_update(right, child_level + 1)

    def generalize(self, k):
        """
        Return the generalization (1 level up in the hierarchy) of the item k
        """
        level, generalization = self.map.get(k, (0, None))
        return generalization

    def extend_itemset(self, it, all_levels = False):
        """
        Extend itemset generalizing each item 1 time
        """
        res = list(it)
        for item in it:
            if all_levels:
                x = self.get_all_generalizations(item)
            else:
                x = self.generalize(item)
            res.extend(x)
        return res
            
    def get_all_generalizations(self, item):
        """
        an iterator over  all the generalizations of item
        """
        item = self.generalize(item)
        while item is not None:
            yield item
            item = self.generalize(item)

    def generalize_itemset(self, itemset, all_levels = False):
        """
        A faster version of generalize_itemset.
        """
        l = len(itemset)
        accumulator = [ [] for x in xrange(l) ]
        for n, item in enumerate(itemset):
            accumulator[n].append(item)
            if all_levels:
                accumulator[n].extend(self.get_all_generalizations(item))
            else:
                upper = self.generalize(item)
                if upper is not None:
                    accumulator[n].append(upper)
        for new_i in cartesian_product(*accumulator):
            f = frozenset(new_i)
            if f != itemset:
                yield f


    def _generalize_itemset(self, itemset, all_levels = False):
        """
        A dumb version of generalize_itemset
        """
        def unique_tags(x_it):
            lx = len(x_it)
            l_diff_tags = len(set(x[0] for x in x_it))
            return lx == l_diff_tags

        l = len(itemset)
        all_generalizations = self.extend_itemset(itemset, all_levels = all_levels)

        new_itemset = [None for x in range(l)]

        for indexes in k_subsets_i(len(all_generalizations), len(itemset)):
            c = 0
            for i in indexes:
                new_itemset[c] = all_generalizations[i]
                c += 1
            new_i = frozenset(new_itemset)
            if new_i != itemset and unique_tags(new_itemset):
                yield new_i

class WellFormedGeneralizer(Generalizer):

    def get_itemset_levels(self, itemset):
        return (self.get_item_level(item) for item in itemset)

    def get_item_level(self, item):
        return self.map.get(item, (0,None))[0]

    def generalize_itemset(self, itemset, all_levels = False, yield_level = False):
        for i in super(WellFormedGeneralizer, self).generalize_itemset(itemset, all_levels = all_levels):
            max_l, can_rise, wf = self.is_well_formed(i)
            if wf:
                if not yield_level:
                    yield i
                else:
                    yield i, max_l

    def is_well_formed(self, itemset):
        """
        Return a tuple max level (int), level can rise (bool), is well formed (bool)
        """
        # Una singola tupla-item e' well-formed
        if isinstance(itemset, tuple):
            level, generalization = self.map.get(itemset, (0, None) )
            return level, generalization is not None,  True

        max_level = -1
        max_level_can_rise = False
        for item in itemset:
            item_level, generalization = self.map.get(item, (0,None))
            if item_level > max_level:
                max_level = item_level
                max_level_can_rise = generalization is not None
        
        for item in itemset:
            item_level, generalization = self.map.get(item, (0,None))
            if item_level < max_level and generalization is not None:
                return max_level, max_level_can_rise, False
        
        return max_level, max_level_can_rise, True

    def get_well_formed_versions(self, itemset):
        max_level, can_rise, is_wf = self.is_well_formed(itemset)
        new_itemset = list(itemset)

        while not is_wf:
            for n, item in enumerate(new_itemset):
                level, generalization = self.map.get(item, (0, None) )
                if level < max_level and generalization is not None:
                    new_itemset[n] = generalization
                else:
                    new_itemset[n] = item
            max_level, can_rise, is_wf = self.is_well_formed(new_itemset)
        
        yield frozenset(new_itemset)

def get_generalizer(algorithm, rules_filename):
    if algorithm == "well-formed" or algorithm == "genio":
        return WellFormedGeneralizer(filename=rules_filename)
    elif "traditional" in algorithm:
        return Generalizer(filename=rules_filename)
    return None
