"""
Run the complete app

main returns the CPU time for the extraction
"""

import tempfile
import os
import shutil

import encode
import io
import sys
import generalize
import time
import timeit

from storage import Storage
from candidate_generation_strategies import get_candidate_generation_procedure

def _main(info_logger, temporary_dir, dataset, minsup, rules_file,
          output_file, optimize = 0, 
          candidate_generation_strategy = None,
          extraction_strategy = None,
          ensure_subsets = False, print_excluded_strategy = None, time_fim = False):
    
    dct_prefix = os.path.join(temporary_dir, "dictionary")
    # Encode dataset
    dataset_encoded = os.path.join(temporary_dir, "%s_encoded" % (os.path.basename(dataset), ))
    info_logger.info("Encoding dataset: file %s on %s" % (dataset, dataset_encoded) )
    
    encode.main(dataset, dct_prefix, dataset_encoded, decode = False, file_type=0, load_old_data=False)
    
    # Encode rules
    rules_encoded = None
    if rules_file:
        rules_encoded = os.path.join(temporary_dir, "%s_encoded" % (os.path.basename(rules_file), ))
        info_logger.info("Encoding rules: file %s on %s" % (rules_file, rules_encoded) )
    
        encode.main(rules_file, dct_prefix, rules_encoded, False, file_type=1, load_old_data=True)
    
    # Mining itemsets
    temp_filename = os.path.join(temporary_dir, 'itemsets_encoded')
    
    info_logger.info("Launching fim_all. minsup = %s. Encoded itemsets written on %s" % (minsup, temp_filename) )
    f = file(temp_filename, 'w')
    
    if "opportunistic" not in extraction_strategy:
        output_writer = io.get_output_writer(f)
    else:
        output_writer = io.get_output_writer(f, print_excluded_strategy = print_excluded_strategy)
    
    no_gc = False
    if optimize > 0:
        no_gc = True
    
    # Funzioni per generalizzare
    generalizer = generalize.get_generalizer(candidate_generation_strategy, rules_filename=rules_encoded)
    
    if extraction_strategy in ["opportunistic", 'opportunistic_basic_pruning']:
        # Generatore di candidati
        from algorithms.opportunistic import genio as fim_all
        candidate_generator = get_candidate_generation_procedure(candidate_generation_strategy, generalizer, extended=True)
    elif extraction_strategy == "opportunistic_post_pruning":
        from algorithms.opportunistic import genio_post_pruning as fim_all
        candidate_generator = get_candidate_generation_procedure(candidate_generation_strategy, generalizer, extended=True)
    elif extraction_strategy == 'sac':
        from algorithms.opportunistic import sac_version as fim_all
        candidate_generator = get_candidate_generation_procedure(candidate_generation_strategy, generalizer, extended=True)
    elif extraction_strategy in ["traditional", "traditional_basic_pruning"]:
        from algorithms.exhaustive import main as fim_all
        # Generatore di candidati
        candidate_generator = get_candidate_generation_procedure(candidate_generation_strategy, generalizer, \
                                                                 exploit_structured = True)
#not (extraction_strategy == "traditional") )
    else:
        raise ValueError, "Extraction strategy '%s' unknown" % (extraction_strategy, )
    
    # Lettura delle transazioni
    dataset_reader = io.DatasetReader(dataset_encoded, generalizer, info_logger.info)
    
    # Prendo i tempi?
    before_fim, after_fim = 0,0
    if not time_fim:
        fim_all(dataset_reader, generalizer, minsup, info_logger,
                output_writer, candidate_generator, no_gc = no_gc, 
                ensure_subsets = ensure_subsets,
                basic_pruning = ('basic_pruning' in extraction_strategy),
                no_exclusion_set = (candidate_generation_strategy == 'traditional_all_generalizations'))
        exe_time = 0
    else:
        import resource
        before_fim = resource.getrusage(resource.RUSAGE_SELF)
        fim_all(dataset_reader, generalizer, minsup, info_logger,
                output_writer, candidate_generator, no_gc = no_gc, 
                ensure_subsets = ensure_subsets,
                basic_pruning = ('basic_pruning' in extraction_strategy),
                no_exclusion_set = (candidate_generation_strategy == 'traditional_all_generalizations'))
        resource.getrusage(resource.RUSAGE_SELF)
        after_fim = resource.getrusage(resource.RUSAGE_SELF)
    
    f.close()
    
    # Decode itemsets
    info_logger.info("Decoding itemsets" )
    itemsets_decoded = os.path.join(temporary_dir, "itemsets_decoded")
    encode.main(temp_filename, dct_prefix, itemsets_decoded, decode = True, file_type=2, load_old_data=True)
    
    # Print output
    if output_file is None:
        info_logger.info("Printing output to standard output" )
        for line in file(itemsets_decoded):
            print line,
    else:
        info_logger.info("Printing output to %s" % (output_file ,) )
        shutil.copyfile(itemsets_decoded, output_file)
    
    return before_fim, after_fim

def main(dataset, minsup, log_file, rules_file, output_file, **kwargs):
    return io.wrap_passing_info_logger_and_tempdir(log_file, _main, dataset, minsup, rules_file, output_file, **kwargs)
